package main

import (
	"bufio"
	"context"
	"dehub/cmd/dehub/dcmd"
	"errors"
	"fmt"
	"io"
	"os"
	"strings"

	"gopkg.in/src-d/go-git.v4/plumbing"
)

func cmdHook(ctx context.Context, cmd *dcmd.Cmd) {
	flag := cmd.FlagSet()
	preRcv := flag.Bool("pre-receive", false, "Use dehub as a server-side pre-receive hook")

	cmd.Run(func() (context.Context, error) {

		if !*preRcv {
			return nil, errors.New("must set the hook type")
		}

		repo := ctxRepo(ctx)

		br := bufio.NewReader(os.Stdin)
		for {
			line, err := br.ReadString('\n')
			if errors.Is(err, io.EOF) {
				break
			} else if err != nil {
				return nil, fmt.Errorf("error reading next line from stdin: %w", err)
			}
			fmt.Printf("Processing line %q\n", strings.TrimSpace(line))

			lineParts := strings.Fields(line)
			if len(lineParts) < 3 {
				return nil, fmt.Errorf("malformed pre-receive hook stdin line %q", line)
			}

			startRev := plumbing.Revision(lineParts[0])
			endRev := plumbing.Revision(lineParts[1])
			branchName := plumbing.ReferenceName(lineParts[2])

			if !branchName.IsBranch() {
				return nil, fmt.Errorf("reference %q is not a branch, can't push to it", branchName)
			}

			gitCommits, err := repo.GetGitRevisionRange(startRev, endRev)
			if err != nil {
				return nil, fmt.Errorf("getting commits from %q to %q: %w",
					startRev, endRev, err)

			} else if err := repo.VerifyCommits(branchName, gitCommits); err != nil {
				return nil, fmt.Errorf("verifying commits from %q to %q: %w",
					startRev, endRev, err)
			}
		}

		fmt.Println("All pushed commits have been verified, well done.")
		return nil, nil
	})
}
